@@ -2,6 +2,8 @@
 
 /*
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2003 Kim Woelders
+ * Copyright (C) 2003 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +25,7 @@
 #include "window.h"
 #include "workspace.h"
 #include "application.h"
+#include "class-group.h"
 #include "xutils.h"
 #include "private.h"
 #include <gdk/gdk.h>
@@ -37,8 +40,7 @@ struct _WnckScreenPrivate
   int number;
   Window xroot;
   Screen *xscreen;
-  int width;
-  int height;
+  
   /* in map order */
   GList *mapped_windows;
   /* in stacking order */
@@ -46,9 +48,18 @@ struct _WnckScreenPrivate
   /* in 0-to-N order */
   GList *workspaces;
 
+  /* previously_active_window is used in tandem with active_window to
+   * determine return status of meta_window_is_most_recently_actived().
+   */
   WnckWindow *active_window;
+  WnckWindow *previously_active_window;
   WnckWorkspace *active_workspace;
 
+  /* Provides the sorting order number for the next window, to make
+   * sure windows remain sorted in the order they appear.
+   */
+  gint window_order;
+
   Pixmap bg_pixmap;
   
   guint update_handler;  
@@ -64,6 +75,7 @@ struct _WnckScreenPrivate
    */
   guint need_update_stack_list : 1;
   guint need_update_workspace_list : 1;
+  guint need_update_viewport_settings : 1;
   guint need_update_active_workspace : 1;
   guint need_update_active_window : 1;
   guint need_update_workspace_names : 1;
@@ -81,8 +93,11 @@ enum {
   WORKSPACE_DESTROYED,
   APPLICATION_OPENED,
   APPLICATION_CLOSED,
+  CLASS_GROUP_OPENED,
+  CLASS_GROUP_CLOSED,
   BACKGROUND_CHANGED,
   SHOWING_DESKTOP_CHANGED,
+  VIEWPORTS_CHANGED,
   LAST_SIGNAL
 };
 
@@ -90,12 +105,13 @@ static void wnck_screen_init        (WnckScreen      *screen);
 static void wnck_screen_class_init  (WnckScreenClass *klass);
 static void wnck_screen_finalize    (GObject         *object);
 
-static void update_client_list      (WnckScreen      *screen);
-static void update_workspace_list   (WnckScreen      *screen);
-static void update_active_workspace (WnckScreen      *screen);
-static void update_active_window    (WnckScreen      *screen);
-static void update_workspace_names  (WnckScreen      *screen);
-static void update_showing_desktop  (WnckScreen      *screen);
+static void update_client_list        (WnckScreen      *screen);
+static void update_workspace_list     (WnckScreen      *screen);
+static void update_viewport_settings  (WnckScreen      *screen);
+static void update_active_workspace   (WnckScreen      *screen);
+static void update_active_window      (WnckScreen      *screen);
+static void update_workspace_names    (WnckScreen      *screen);
+static void update_showing_desktop    (WnckScreen      *screen);
 
 static void queue_update            (WnckScreen      *screen);
 static void unqueue_update          (WnckScreen      *screen);              
@@ -116,8 +132,13 @@ static void emit_application_opened       (WnckScreen      *screen,
                                            WnckApplication *app);
 static void emit_application_closed       (WnckScreen      *screen,
                                            WnckApplication *app);
+static void emit_class_group_opened       (WnckScreen      *screen,
+                                           WnckClassGroup  *class_group);
+static void emit_class_group_closed       (WnckScreen      *screen,
+                                           WnckClassGroup  *class_group);
 static void emit_background_changed       (WnckScreen      *screen);
 static void emit_showing_desktop_changed  (WnckScreen      *screen);
+static void emit_viewports_changed        (WnckScreen      *screen);
 
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -251,6 +272,34 @@ wnck_screen_class_init (WnckScreenClass *klass)
                   g_cclosure_marshal_VOID__OBJECT,
                   G_TYPE_NONE, 1, WNCK_TYPE_APPLICATION);
 
+  signals[CLASS_GROUP_OPENED] =
+    g_signal_new ("class_group_opened",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+#if 0
+                  /* FIXME when we can break ABI add this */
+                  G_STRUCT_OFFSET (WnckScreenClass, class_group_opened),
+#else
+		  0,
+#endif
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE, 1, WNCK_TYPE_CLASS_GROUP);
+
+  signals[CLASS_GROUP_CLOSED] =
+    g_signal_new ("class_group_closed",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+#if 0
+                  /* FIXME when we can break ABI add this */
+                  G_STRUCT_OFFSET (WnckScreenClass, class_group_closed),
+#else
+		  0,
+#endif
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE, 1, WNCK_TYPE_CLASS_GROUP);
+
   signals[BACKGROUND_CHANGED] =
     g_signal_new ("background_changed",
                   G_OBJECT_CLASS_TYPE (object_class),
@@ -268,6 +317,20 @@ wnck_screen_class_init (WnckScreenClass *klass)
                   G_STRUCT_OFFSET (WnckScreenClass, showing_desktop_changed),
 #else
                   0,
+#endif
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
+
+    signals[VIEWPORTS_CHANGED] =
+    g_signal_new ("viewports_changed",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+#if 0
+                  /* FIXME when we can break ABI add this */
+                  G_STRUCT_OFFSET (WnckScreenClass, viewports_changed),
+#else
+                  0,
 #endif
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
@@ -326,6 +389,7 @@ wnck_screen_construct (WnckScreen *screen,
   screen->priv->xroot = RootWindow (gdk_display, number);
   screen->priv->xscreen = ScreenOfDisplay (gdk_display, number);
   screen->priv->number = number;
+  screen->priv->window_order = 0;
 
 #ifdef HAVE_STARTUP_NOTIFICATION
   screen->priv->sn_display = sn_display_new (gdk_display,
@@ -338,11 +402,9 @@ wnck_screen_construct (WnckScreen *screen,
   _wnck_select_input (screen->priv->xroot,
                       PropertyChangeMask);
 
-  screen->priv->width = WidthOfScreen (screen->priv->xscreen);
-  screen->priv->height = HeightOfScreen (screen->priv->xscreen);
-
   screen->priv->need_update_workspace_list = TRUE;
   screen->priv->need_update_stack_list = TRUE;
+  screen->priv->need_update_viewport_settings = TRUE;
   screen->priv->need_update_active_workspace = TRUE;
   screen->priv->need_update_active_window = TRUE;
   screen->priv->need_update_workspace_names = TRUE;
@@ -498,6 +560,24 @@ wnck_screen_get_active_window (WnckScreen *screen)
   return screen->priv->active_window;
 }
 
+/**
+ * wnck_screen_get_previously_active_window:
+ * @screen: a #WnckScreen
+ * 
+ * Gets the previously active window. May return %NULL
+ * sometimes, since not all window managers guarantee
+ * that a window is always active.
+ * 
+ * Return value: previously active window or %NULL
+ **/
+WnckWindow*
+wnck_screen_get_previously_active_window (WnckScreen *screen)
+{
+  g_return_val_if_fail (WNCK_IS_SCREEN (screen), NULL);
+
+  return screen->priv->previously_active_window;
+}
+
 /**
  * wnck_screen_get_windows:
  * @screen: a #WnckScreen
@@ -587,6 +667,38 @@ wnck_screen_get_workspace_count (WnckScreen *screen)
   return g_list_length (screen->priv->workspaces);
 }
 
+/**
+ * wnck_screen_change_workspace_count:
+ * @screen: a #WnckScreen
+ * @count: requested count
+ * 
+ * Asks the window manager to change the number of workspaces.
+ **/
+void
+wnck_screen_change_workspace_count (WnckScreen *screen,
+                                    int         count)
+{
+  XEvent xev;
+  
+  g_return_if_fail (WNCK_IS_SCREEN (screen));
+  g_return_if_fail (count >= 1);
+  
+  xev.xclient.type = ClientMessage;
+  xev.xclient.serial = 0;
+  xev.xclient.window = screen->priv->xroot;
+  xev.xclient.send_event = True;
+  xev.xclient.display = DisplayOfScreen (screen->priv->xscreen);
+  xev.xclient.message_type = _wnck_atom_get ("_NET_NUMBER_OF_DESKTOPS");
+  xev.xclient.format = 32;
+  xev.xclient.data.l[0] = count;
+  
+  XSendEvent (DisplayOfScreen (screen->priv->xscreen),
+              screen->priv->xroot,
+              False,
+              SubstructureRedirectMask | SubstructureNotifyMask,
+              &xev);
+}
+
 void
 _wnck_screen_process_property_notify (WnckScreen *screen,
                                       XEvent     *xevent)
@@ -612,6 +724,18 @@ _wnck_screen_process_property_notify (WnckScreen *screen,
       screen->priv->need_update_stack_list = TRUE;
       queue_update (screen);
     }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("_NET_DESKTOP_VIEWPORT"))
+    {
+      screen->priv->need_update_viewport_settings = TRUE;
+      queue_update (screen);
+    }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("_NET_DESKTOP_GEOMETRY"))
+    {
+      screen->priv->need_update_viewport_settings = TRUE;
+      queue_update (screen);
+    }
   else if (xevent->xproperty.atom ==
            _wnck_atom_get ("_NET_NUMBER_OF_DESKTOPS"))
     {
@@ -725,8 +849,8 @@ update_client_list (WnckScreen *screen)
   GList *new_list;
   GList *created;
   GList *closed;
-  GList *created_apps;
-  GList *closed_apps;
+  GList *created_apps, *closed_apps;
+  GList *created_class_groups, *closed_class_groups;
   GList *tmp;
   int i;
   GHashTable *new_hash;
@@ -768,25 +892,34 @@ update_client_list (WnckScreen *screen)
   closed = NULL;
   created_apps = NULL;
   closed_apps = NULL;
+  created_class_groups = NULL;
+  closed_class_groups = NULL;
 
   new_hash = g_hash_table_new (NULL, NULL);
   
-  new_stack_list = NULL;
+  new_list = NULL;
   i = 0;
-  while (i < stack_length)
+  while (i < mapping_length)
     {
       WnckWindow *window;
 
-      window = wnck_window_get (stack[i]);
+      window = wnck_window_get (mapping[i]);
 
       if (window == NULL)
         {
           Window leader;
           WnckApplication *app;
+	  const char *res_class;
+	  WnckClassGroup *class_group;
           
-          window = _wnck_window_create (stack[i], screen);
+          window = _wnck_window_create (mapping[i], 
+                                        screen,
+                                        screen->priv->window_order++);
+
           created = g_list_prepend (created, window);
 
+	  /* Application */
+
           leader = wnck_window_get_group_leader (window);
           
           app = wnck_application_get (leader);
@@ -797,9 +930,22 @@ update_client_list (WnckScreen *screen)
             }
           
           _wnck_application_add_window (app, window);
+
+	  /* Class group */
+
+	  res_class = _wnck_window_get_resource_class (window);
+
+	  class_group = wnck_class_group_get (res_class);
+	  if (class_group == NULL)
+	    {
+	      class_group = _wnck_class_group_create (res_class);
+	      created_class_groups = g_list_prepend (created_class_groups, class_group);
+	    }
+
+	  _wnck_class_group_add_window (class_group, window);
         }
 
-      new_stack_list = g_list_prepend (new_stack_list, window);
+      new_list = g_list_prepend (new_list, window);
 
       g_hash_table_insert (new_hash, window, window);
       
@@ -807,12 +953,12 @@ update_client_list (WnckScreen *screen)
     }
       
   /* put list back in order */
-  new_stack_list = g_list_reverse (new_stack_list);
+  new_list = g_list_reverse (new_list);
 
   /* Now we need to find windows in the old list that aren't
    * in this new list
    */
-  tmp = screen->priv->stacked_windows;
+  tmp = screen->priv->mapped_windows;
   while (tmp != NULL)
     {
       WnckWindow *window = tmp->data;
@@ -820,15 +966,25 @@ update_client_list (WnckScreen *screen)
       if (g_hash_table_lookup (new_hash, window) == NULL)
         {
           WnckApplication *app;
+	  WnckClassGroup *class_group;
           
           closed = g_list_prepend (closed, window);
 
-          app = wnck_window_get_application (window);
+	  /* Remove from the app */
 
+          app = wnck_window_get_application (window);
           _wnck_application_remove_window (app, window);
 
           if (wnck_application_get_windows (app) == NULL)
             closed_apps = g_list_prepend (closed_apps, app);
+
+	  /* Remove from the class group */
+
+          class_group = wnck_window_get_class_group (window);
+          _wnck_class_group_remove_window (class_group, window);
+
+          if (wnck_class_group_get_windows (class_group) == NULL)
+            closed_class_groups = g_list_prepend (closed_class_groups, class_group);
         }
       
       tmp = tmp->next;
@@ -837,17 +993,17 @@ update_client_list (WnckScreen *screen)
   g_hash_table_destroy (new_hash);
 
   /* Now get the mapping in list form */
-  new_list = NULL;
+  new_stack_list = NULL;
   i = 0;
-  while (i < mapping_length)
+  while (i < stack_length)
     {
       WnckWindow *window;
 
-      window = wnck_window_get (mapping[i]);
+      window = wnck_window_get (stack[i]);
 
       g_assert (window != NULL);
 
-      new_list = g_list_prepend (new_list, window);
+      new_stack_list = g_list_prepend (new_stack_list, window);
       
       ++i;
     }
@@ -856,7 +1012,7 @@ update_client_list (WnckScreen *screen)
   g_free (mapping);
       
   /* put list back in order */
-  new_list = g_list_reverse (new_list);
+  new_stack_list = g_list_reverse (new_stack_list);
   
   /* Now new_stack_list becomes screen->priv->stack_windows, new_list
    * becomes screen->priv->mapped_windows, and we emit the opened/closed
@@ -872,6 +1028,8 @@ update_client_list (WnckScreen *screen)
       g_assert (closed == NULL);
       g_assert (created_apps == NULL);
       g_assert (closed_apps == NULL);
+      g_assert (created_class_groups == NULL);
+      g_assert (closed_class_groups == NULL);
       g_list_free (new_stack_list);
       g_list_free (new_list);      
       --reentrancy_guard;
@@ -879,7 +1037,7 @@ update_client_list (WnckScreen *screen)
     }
 
   g_list_free (screen->priv->mapped_windows);
-  g_list_free (screen->priv->stacked_windows);  
+  g_list_free (screen->priv->stacked_windows);
   screen->priv->mapped_windows = new_list;
   screen->priv->stacked_windows = new_stack_list;
 
@@ -888,31 +1046,22 @@ update_client_list (WnckScreen *screen)
    * don't handle it, but we do warn about it using reentrancy_guard
    */
 
-  /* Sequence is: application_opened, window_opened, window_closed,
-   * application_closed. We have to do all window list changes
-   * BEFORE doing any other signals, so that any observers
-   * have valid state for the window structure before they take
-   * further action
+  /* Sequence is: class_group_opened, application_opened, window_opened,
+   * window_closed, application_closed, class_group_closed. We have to do all
+   * window list changes BEFORE doing any other signals, so that any observers
+   * have valid state for the window structure before they take further action
    */
-  tmp = created_apps;
-  while (tmp != NULL)
-    {
-      emit_application_opened (screen, WNCK_APPLICATION (tmp->data));
-      
-      tmp = tmp->next;
-    }
-  
-  tmp = created;
-  while (tmp != NULL)
-    {
-      emit_window_opened (screen, WNCK_WINDOW (tmp->data));
-      
-      tmp = tmp->next;
-    }
+  for (tmp = created_class_groups; tmp; tmp = tmp->next)
+    emit_class_group_opened (screen, WNCK_CLASS_GROUP (tmp->data));
+
+  for (tmp = created_apps; tmp; tmp = tmp->next)
+    emit_application_opened (screen, WNCK_APPLICATION (tmp->data));
+
+  for (tmp = created; tmp; tmp = tmp->next)
+    emit_window_opened (screen, WNCK_WINDOW (tmp->data));
 
   active_changed = FALSE;
-  tmp = closed;
-  while (tmp != NULL)
+  for (tmp = closed; tmp; tmp = tmp->next)
     {
       WnckWindow *window;
 
@@ -920,26 +1069,24 @@ update_client_list (WnckScreen *screen)
 
       if (window == screen->priv->active_window)
         {
+          screen->priv->previously_active_window = screen->priv->active_window;
           screen->priv->active_window = NULL;
           active_changed = TRUE;
         }
+
+      if (window == screen->priv->previously_active_window)
+        {
+          screen->priv->previously_active_window = NULL;
+        }
       
       emit_window_closed (screen, window);
-      
-      tmp = tmp->next;
     }
 
-  tmp = closed_apps;
-  while (tmp != NULL)
-    {
-      WnckApplication *app;
+  for (tmp = closed_apps; tmp; tmp = tmp->next)
+    emit_application_closed (screen, WNCK_APPLICATION (tmp->data));
 
-      app = WNCK_APPLICATION (tmp->data);
-
-      emit_application_closed (screen, app);
-      
-      tmp = tmp->next;
-    }
+  for (tmp = closed_class_groups; tmp; tmp = tmp->next)
+    emit_class_group_closed (screen, WNCK_CLASS_GROUP (tmp->data));
 
   if (stack_changed)
     emit_window_stacking_changed (screen);
@@ -948,29 +1095,17 @@ update_client_list (WnckScreen *screen)
     emit_active_window_changed (screen);
   
   /* Now free the closed windows */
-  tmp = closed;
-  while (tmp != NULL)
-    {
-      WnckWindow *window = tmp->data;
-
-      _wnck_window_destroy (window);
-      
-      tmp = tmp->next;
-    }
+  for (tmp = closed; tmp; tmp = tmp->next)
+    _wnck_window_destroy (WNCK_WINDOW (tmp->data));
 
   /* Free the closed apps */
-  tmp = closed_apps;
-  while (tmp != NULL)
-    {
-      WnckApplication *app;
+  for (tmp = closed_apps; tmp; tmp = tmp->next)
+    _wnck_application_destroy (WNCK_APPLICATION (tmp->data));
 
-      app = WNCK_APPLICATION (tmp->data);
+  /* Free the closed class groups */
+  for (tmp = closed_class_groups; tmp; tmp = tmp->next)
+    _wnck_class_group_destroy (WNCK_CLASS_GROUP (tmp->data));
 
-      _wnck_application_destroy (app);
-      
-      tmp = tmp->next;
-    }
-  
   g_list_free (closed);
   g_list_free (created);
   g_list_free (closed_apps);
@@ -1105,6 +1240,118 @@ update_workspace_list (WnckScreen *screen)
   --reentrancy_guard;
 }
 
+static void
+update_viewport_settings (WnckScreen *screen)
+{
+  int i, n_spaces;
+  WnckWorkspace *space;
+  gulong *p_coord;
+  int n_coord;
+  gboolean do_update;
+  int space_width, space_height;
+  gboolean got_viewport_prop;
+  
+  if (!screen->priv->need_update_viewport_settings)
+    return;
+
+  screen->priv->need_update_viewport_settings = FALSE;
+
+  do_update = FALSE;
+
+  n_spaces = wnck_screen_get_workspace_count (screen);
+
+  /* If no property, use the screen's size */
+  space_width = wnck_screen_get_width (screen);
+  space_height = wnck_screen_get_height (screen);
+  
+  if (_wnck_get_cardinal_list (screen->priv->xroot,
+			       _wnck_atom_get ("_NET_DESKTOP_GEOMETRY"),
+                               &p_coord, &n_coord) &&
+      p_coord != NULL)
+    {
+      if (n_coord == 2)
+	{
+          space_width = p_coord[0];
+          space_height = p_coord[1];
+          
+          if (space_width < wnck_screen_get_width (screen))
+            space_width = wnck_screen_get_width (screen);
+
+          if (space_height < wnck_screen_get_height (screen))
+            space_height = wnck_screen_get_height (screen);
+	}
+      
+      g_free (p_coord);
+    }
+          
+  for (i = 0; i < n_spaces; i++)
+    {
+      space = wnck_screen_get_workspace (screen, i);
+      g_assert (space != NULL);
+      
+      if (_wnck_workspace_set_geometry (space, space_width, space_height))
+        do_update = TRUE;
+    }
+
+  got_viewport_prop = FALSE;
+  
+  if (_wnck_get_cardinal_list (screen->priv->xroot,
+                               _wnck_atom_get ("_NET_DESKTOP_VIEWPORT"),
+                               &p_coord, &n_coord) &&
+      p_coord != NULL)
+    {
+      if (n_coord == 2 * n_spaces)
+        {
+          int screen_width, screen_height;
+
+          got_viewport_prop = TRUE;
+          
+          screen_width = wnck_screen_get_width (screen);
+          screen_height = wnck_screen_get_height (screen);
+          
+	  for (i = 0; i < n_spaces; i++)
+	    {
+              int x = 2 * i;
+              int y = 2 * i + 1;
+
+              space = wnck_screen_get_workspace (screen, i);
+              g_assert (space != NULL);
+              
+              if (p_coord[x] < 0)
+                p_coord[x] = 0;
+              else if (p_coord[x] > space_width - screen_width)
+                p_coord[x] = space_width - screen_width;
+
+              if (p_coord[y] < 0)
+                p_coord[y] = 0;
+              else if (p_coord[y] > space_height - screen_height)
+                p_coord[y] = space_height - screen_height;
+
+	      if (_wnck_workspace_set_viewport (space,
+                                                p_coord[x], p_coord[y]))
+                do_update = TRUE;
+	    }
+	}
+      
+      g_free (p_coord);
+    }
+
+  if (!got_viewport_prop)
+    {
+      for (i = 0; i < n_spaces; i++)
+        {
+          space = wnck_screen_get_workspace (screen, i);
+          g_assert (space != NULL);
+          
+          if (_wnck_workspace_set_viewport (space, 0, 0))
+            do_update = TRUE;
+        }
+    }
+  
+  if (do_update)
+    emit_viewports_changed (screen);
+}
+
 static void
 update_active_workspace (WnckScreen *screen)
 {
@@ -1153,6 +1400,7 @@ update_active_window (WnckScreen *screen)
   if (window == screen->priv->active_window)
     return;
 
+  screen->priv->previously_active_window = screen->priv->active_window;
   screen->priv->active_window = window;
 
   emit_active_window_changed (screen);
@@ -1192,6 +1440,7 @@ update_workspace_names (WnckScreen *screen)
     }
 
   g_strfreev (names);
+
   g_list_free (copy);
 }
 
@@ -1245,12 +1494,24 @@ do_update_now (WnckScreen *screen)
       screen->priv->update_handler = 0;
     }
 
+  /* if number of workspaces changes, we have to
+   * update the per-workspace information as well
+   * in case the WM changed the per-workspace info
+   * first and number of spaces second.
+   */
+  if (screen->priv->need_update_workspace_list)
+    {
+      screen->priv->need_update_viewport_settings = TRUE;
+      screen->priv->need_update_workspace_names = TRUE;
+    }
+      
   /* First get our big-picture state in order */
   update_workspace_list (screen);
   update_client_list (screen);
 
   /* Then note any smaller-scale changes */
   update_active_workspace (screen);
+  update_viewport_settings (screen);
   update_active_window (screen);
   update_workspace_names (screen);
   update_showing_desktop (screen);
@@ -1369,6 +1630,24 @@ emit_application_closed (WnckScreen      *screen,
                  0, app);
 }
 
+static void
+emit_class_group_opened (WnckScreen     *screen,
+                         WnckClassGroup *class_group)
+{
+  g_signal_emit (G_OBJECT (screen),
+                 signals[CLASS_GROUP_OPENED],
+                 0, class_group);
+}
+
+static void
+emit_class_group_closed (WnckScreen     *screen,
+                         WnckClassGroup *class_group)
+{
+  g_signal_emit (G_OBJECT (screen),
+                 signals[CLASS_GROUP_CLOSED],
+                 0, class_group);
+}
+
 static void
 emit_background_changed (WnckScreen *screen)
 {
@@ -1385,6 +1664,14 @@ emit_showing_desktop_changed (WnckScreen *screen)
                  0);
 }
 
+static void
+emit_viewports_changed (WnckScreen *screen)
+{
+  g_signal_emit (G_OBJECT (screen),
+                 signals[VIEWPORTS_CHANGED],
+                 0);
+}
+
 gboolean
 wnck_screen_net_wm_supports (WnckScreen *screen,
                              const char *atom)
@@ -1405,7 +1692,7 @@ wnck_screen_get_width (WnckScreen *screen)
 {
   g_return_val_if_fail (WNCK_IS_SCREEN (screen), 0);
 
-  return gdk_screen_width ();
+  return WidthOfScreen (screen->priv->xscreen);
 }
 
 int
@@ -1413,7 +1700,7 @@ wnck_screen_get_height (WnckScreen *screen)
 {
   g_return_val_if_fail (WNCK_IS_SCREEN (screen), 0);
 
-  return gdk_screen_height ();
+  return HeightOfScreen (screen->priv->xscreen);
 }
 
 Screen *
@@ -1476,6 +1763,27 @@ wnck_screen_toggle_showing_desktop (WnckScreen *screen,
                                 show);
 }
 
+
+/**
+ * wnck_screen_move_viewport:
+ * @screen: a #WnckScreen
+ * @x: X offset of viewport
+ * @y: Y offset of viewport
+ *
+ * Ask window manager to move the viewport of the current workspace.
+ */
+void
+wnck_screen_move_viewport (WnckScreen *screen,
+                           int         x,
+                           int         y)
+{
+  g_return_if_fail (WNCK_IS_SCREEN (screen));
+  g_return_if_fail (x >= 0);
+  g_return_if_fail (y >= 0);
+  
+  _wnck_change_viewport (WNCK_SCREEN_XSCREEN (screen), x, y);
+}
+
 #ifdef HAVE_STARTUP_NOTIFICATION
 SnDisplay*
 _wnck_screen_get_sn_display (WnckScreen *screen)
@@ -1485,3 +1793,41 @@ _wnck_screen_get_sn_display (WnckScreen *screen)
   return screen->priv->sn_display;
 }
 #endif /* HAVE_STARTUP_NOTIFICATION */
+
+void
+_wnck_screen_change_workspace_name (WnckScreen *screen,
+                                    int         number,
+                                    const char *name)
+{
+  int n_spaces;
+  char **names;
+  int i;
+  
+  n_spaces = wnck_screen_get_workspace_count (screen);
+
+  names = g_new0 (char*, n_spaces + 1);
+
+  i = 0;
+  while (i < n_spaces)
+    {
+      if (i == number)
+        names[i] = (char*) name;
+      else
+        {
+          WnckWorkspace *workspace;
+          workspace = wnck_screen_get_workspace (screen, i);
+          if (workspace)
+            names[i] = (char*) wnck_workspace_get_name (workspace);
+          else
+            names[i] = (char*) ""; /* maybe this should be a g_warning() */
+        }
+      
+      ++i;
+    }
+
+  _wnck_set_utf8_list (screen->priv->xroot,
+                       _wnck_atom_get ("_NET_DESKTOP_NAMES"),
+                       names);
+
+  g_free (names);
+}
